#include "kvlist.h"

kvlist* kvlistCreate()
{
	kvlist* ret;

	ret = (kvlist*)smalloc(sizeof(kvlist));
	ret->head = NULL;
	ret->tail = NULL;
	ret->len = 0;

	return ret;
}

void kvlistDestroy(kvlist* restrict kvl)
{
	assert(kvl != NULL);
	assert(kvl->head == NULL);
	assert(kvl->tail == NULL);
	assert(kvl->len == 0);

	sfree(kvl);
}

void kvlistPush(kvlist* restrict kvl, keyval* kv)
{
	kvlistnode* kvn;

	assert(kvl != NULL);
	assert(kv != NULL);

	kvn = (kvlistnode*)smalloc(sizeof(kvlistnode));
	kvn->kv = kv;
	kvn->next = NULL;

	/* Head/Tail NULL */
	if(kvl->head == NULL && kvl->tail == NULL)
	{
		kvl->head = kvn;
		kvl->tail = kvn;
	}

	/* We have at least one element */
	else
	{
		kvl->tail->next = kvn;
		kvl->tail = kvn;
	}

	(kvl->len)++;
}

keyval* kvlistPop(kvlist* restrict kvl)
{
	kvlistnode* cur;
	kvlistnode* tmp;
	keyval* ret;

	cur = kvl->head;

	if(cur->next)
	{
		while(cur->next->next != NULL)
		{
			cur = cur->next;
		}
		tmp = cur->next;

		cur->next = NULL;
		kvl->tail = cur;

		ret = tmp->kv;
		sfree(tmp);
	}
	else
	{
		tmp = cur;
		kvl->tail = NULL;
		kvl->head = NULL;

		ret = tmp->kv;
		sfree(tmp);
	}

	(kvl->len)--;

	return ret;
}

inline bool kvlistIsEmpty(kvlist* restrict kvl)
{
	assert(kvl != NULL);
	return (kvl->len == 0);
}

void kvlistOutput(kvlist* restrict kvl)
{
	kvlistnode* cur;

	assert(kvl != NULL);

	cur = kvl->head;
	while(cur)
	{
		printf("Key: %s, Val: %s\n", cur->kv->key, cur->kv->val);
		cur = cur->next;
	}
}
