/* Malloc and such */
#include <stdlib.h>

/* Shared */
#include "shared.h"

/* Need key-value stuff */
#include "keyval.h"

/* printf */
#include <stdio.h>

#ifndef KVLIST_H
#define KVLIST_H

/**
 * Struct for holding a key-value list node
*/
typedef struct kvlistNode {
	keyval* kv;
	struct kvlistNode* next;
} kvlistnode;

/**
 * Struct for holding a key-value list
 */
typedef struct kvList {
	kvlistnode* head;
	kvlistnode* tail;
	unsigned int len;
} kvlist;

/**
 * Creates an empty kvlist struct.
 *
 * @return The kvlist struct that was created.
 */
kvlist* kvlistCreate(void) __attribute__((malloc));

void kvlistDestroy(kvlist* kvl) __attribute__((nonnull));

/**
 * Pushes the given key-value pair pointer to the end of the list
 *
 * @param kv The key-value pair pointer to add to the end of the list
 */
void kvlistPush(kvlist* restrict kvl, keyval* kv) __attribute__((nonnull));

/**
 * Pops the last key-value pair from the end of the list
 *
 * @return The last key-value pair.  MUST BE FREED BY CALLER.
 */
keyval* kvlistPop(kvlist* restrict kvl) __attribute__((nonnull));

inline bool kvlistIsEmpty(kvlist* restrict kvl) __attribute__((nonnull,pure));

void kvlistOutput(kvlist* restrict kvl) __attribute__((nonnull));

#endif /* KVLIST_H */
