#include "map.h"

map* loadMap(const char* restrict filename)
{
	FILE* restrict mapf;
	char buf[80];
	unsigned short int lines = 0;
	unsigned short int i;
	unsigned short int w, h;
	char* t;

	char* first;
	char* second;

	map* restrict ret;
	
	assert(filename != NULL);

	if(!(mapf = fopen(filename, "rb")))
	{
		fprintf(stderr, "loadMap:\tFailed to open map file (%s).\n", filename);
		return NULL;
	}

	ret = (map*)smalloc(sizeof(map));

	while(fgets(buf, 80, mapf))
	{
		if(lines < 3)
		{
			first = buf;
			second = strchr(buf, 'x') + 1;

			/* If we're reading in the pixel start location */
			if(lines == 0)
			{
				ret->startx = atoi(first);
				ret->starty = atoi(second);
			}

			/* If we're reading the dimension for tiles */
			else if(lines == 1)
			{
				ret->tilex = atoi(first);
				ret->tiley = atoi(second);
			}

			/* If we're reading the dimention of the map in tiles */
			else
			{
				ret->maptx = atoi(first);
				ret->mapty = atoi(second);
			}
		}
		/* If we're reading a filename */
		else
		{
			//TODO: Handle case where a tile doesn't exist.
			if(lines == 3)
			{
				ret->tiles = (char***)calloc(ret->maptx, sizeof(char**));
				for(i = 0; i < ret->maptx; i++)
				{
					ret->tiles[i] = (char**)calloc(ret->mapty, sizeof(char*));
				}

				w = 0;
				h = 0;
			}

			t = strchr(buf, '\n');
			*t = '\0';
			ret->tiles[w][h] = (char*)calloc(strlen(buf)+1, sizeof(char));
			strcpy(ret->tiles[w][h], buf);

			w = (w + 1) % ret->maptx;
			if(w % ret->maptx == 0)
				++h;
		}
		++lines;
	}

	/* We're done reading now */
	fclose(mapf);

	return ret;
}

void freeMap(map* restrict m)
{
	unsigned short int i, j;

	assert(m != NULL);

	for(i = 0; i < m->maptx; i++)
	{
		for(j = 0; j < m->mapty; j++)
		{
			sfree(m->tiles[i][j]);
		}
		sfree(m->tiles[i]);
	}
	sfree(m->tiles);
	sfree(m);
}
