#include "netin.h"

void avatarCopy(avatar* restrict dest, const avatar* restrict src)
{
	dest->x = src->x;			
	dest->y = src->y;
	dest->r = src->r;
	dest->vx = src->vx;
	dest->vy = src->vy;
	dest->score = src->score;
	dest->shape = src->shape;
	dest->popped = src->popped;	
}

int netIn(void* data)
{
	const bool* restrict done = ((bool**)data)[0];
	avatarList* restrict avatars = ((avatarList**)data)[1];
	const TCPsocket sock = *((TCPsocket**)data)[2];
	SDL_mutex* goLock = ((SDL_mutex**)data)[3];
	SDL_mutex* avatarsLock = ((SDL_mutex**)data)[4];

	/* Iteration var */
	unsigned short int i;

	/* The size of an avatar, now we don't have to compute it every time */
	static const unsigned short int szav = sizeof(avatar);

	/* The buffer for gettin' data */
	char buf[szav];

	/* How many bytes were recieved from the server */
	unsigned short int brecv;

	/* The avatar we got over the wire */
	avatar* gotava;

	assert(done != NULL);
	assert(avatars != NULL);
	assert(goLock != NULL);
	assert(avatarsLock != NULL);

	/* While we're not done */
	i = 0;

	SDL_LockMutex(goLock);
	while(!*done)
	{
		/* Clear our buffer */
		memset(buf, 0, szav);

		/* Let's get us some data from the server, you dig? */
		if((brecv = srecv(sock, buf, szav)) != szav)
		{
			fprintf(stderr, "NetIn:\tThe server's supposed to be sendin' stuff %i long, but it's an ass and sent %i\n", szav, brecv); // TODO: Handle this
			fprintf(stderr, "SDL ERROR: %s\n", SDLNet_GetError());
			continue;
		}

		/* Is the server tellin' us to quit? */
		if(unlikely(!strncmp(buf, "QS", 2)))
		{
			/* Okay server, we can do that */
			break;
		}

		/* This is an -avatar-, we can deal with that */
		/* What do we got to do here?
			1) Get the avatar
			2) Is it in our list?
				3) (Ted Stevens) No!  Allocate some space for it then, man.
				3) Yeah.  Update it then!
		*/
		else
		{
			/* Let's make it look like an avatar first */
			gotava = (avatar*)buf;

			/* Do we got it already? */
			if(avatars->as[gotava->id] != NULL)
			{
				/* Update what we got */
				SDL_LockMutex(avatarsLock);
				avatarCopy(avatars->as[gotava->id], gotava);
				SDL_UnlockMutex(avatarsLock);
			}

			/* We need to allocate it and -then- update it! */
			else
			{
				SDL_LockMutex(avatarsLock);
				avatars->as[gotava->id] = (avatar*)malloc(sizeof(avatar));
				avatarCopy(avatars->as[gotava->id], gotava);
				++(avatars->cur);
				SDL_UnlockMutex(avatarsLock);
			}
		}
		/* Let main thread continue */
		if(likely(++i > avatars->max))
		{
			SDL_UnlockMutex(goLock);
		}
	}

	/* Tell the server that we're quitting */
	*buf = 'E';
	*(buf+1) = 'O';
	if(ssend(sock, buf, 2) != 2)
	{
		fprintf(stderr, "Could not tell the server that we're quitting\n");
	}
	sclose(sock);

	/* Free our avatars too */
	for(i = 0; i < avatars->max; i++)
	{
		if(avatars->as[i] != NULL)
		{
			sfree(avatars->as[i]);
		}
	}

	return SUCCESS;
}
