#include "networking.h"

inline void sclose(TCPsocket sock)
{
	SDLNet_TCP_Close(sock);
}

inline ssize_t ssend(TCPsocket sock, const void* restrict data, size_t length)
{
	assert(sock != NULL);
	assert(data != NULL);
	assert(length > 0 && length < 1024);

	return SDLNet_TCP_Send(sock, data, length);
}

inline ssize_t srecv(TCPsocket sock, void* restrict buffer, size_t length)
{
	assert(sock != NULL);
	assert(buffer != NULL);
	assert(length > 0 && length < 1024);

	return SDLNet_TCP_Recv(sock, buffer, length);
}

TCPsocket makeServerSocket(const unsigned short int port)
{
	/* Our own IP Address */
	IPaddress ip;

	/* The socket (this is a pointer) */
	TCPsocket sock;

	if(SDLNet_ResolveHost(&ip, NULL, port) < 0)
	{
		fprintf(stderr, "Networking:\tCould not resolve host while making server socket on port %i\n", port);
		return NULL;
	}

	if((sock = SDLNet_TCP_Open(&ip)) == NULL)
	{
		fprintf(stderr, "Networking:\tCould not get socket while making server socket on port %i\n", port);
		return NULL;
	}

	return sock;
}

TCPsocket acceptConnection(TCPsocket listen, const bool* restrict done)
{
	TCPsocket client;

	while(!*done)
	{
		if((client = SDLNet_TCP_Accept(listen)) == NULL)
		{
			/* There was no connection */
			SDL_Delay(100);
			continue;
		}

		return client;
	}

	return NULL;
}

TCPsocket socketConnect(const char* restrict server, const unsigned short int port)
{
	/* The server's resolved ip address */
	IPaddress ip;

	/* The socket to return */
	TCPsocket sock;

	/* Resolve the hostname into an ip */
	if(SDLNet_ResolveHost(&ip, server, port) < 0)
	{
		fprintf(stderr, "Networking:\tCould not resolve host %s for port %u\n", server, port);
		return NULL;
	}

	/* Open the socket */
	sock = SDLNet_TCP_Open(&ip);
	if(!sock)
	{
		fprintf(stderr, "Networking:\tCould not open connection to %s on port %i\n", server, port);
		return NULL;
	}

	return sock;
}
