#include "servernet.h"

int serverAccept(void* data)
{
	const bool* restrict done = ((bool**)data)[0];
	TCPsocket listenSock = *(((TCPsocket**)data)[1]);
	clientList* restrict clients = ((clientList**)data)[2];
	avatarList* restrict avatars = ((avatarList**)data)[3];

	/* Pointer to a new client when they connect and their avatar */
	client* newClient;
	avatar* newAvatar;

	/* The client's socket */
	TCPsocket sock;

	/* Quadrinary-pointer array for passing done, clientnumber, clients and avatars to thread funcs */
	void* dCAC[4];

	/* Which client this one is in the client array */
	unsigned char clientnumber;

	/* Buffer for sending the client first bit o' infos */
	char sBuf[sizeof(avatar)];

	assert(done != NULL);
	assert(listenSock != NULL);
	assert(clients != NULL);

	/* Initialize as much of dCAC as possible */
	dCAC[0] = (void*)done;
	dCAC[2] = clients;
	dCAC[3] = avatars;

	while(!*done)
	{
		printf("SA:\tWaiting for a client to connect\n");

		/* Get a connection from a client */
		if((sock = acceptConnection(listenSock, done)) == NULL)
		{
			if(*done)
			{
				continue;
			}
			printf("SA:\tERROR: Could not accept connection\n");
			continue;
		}

		/* If we already have the max number of players */
		else if(clients->cur >= clients->max)
		{
			printf("SA:\tServer is full, rejecting client\n");
			if(ssend(sock, "FULL", 4) < 4)
			{
				printf("SA:\tERROR: Could not tell potential client that server is full\n");
			}
			sclose(sock);
		}

		/* Everything looks good */
		else
		{
			/* Get their spot in the client list */
			clientnumber = 0;
			while(clients->cs[clientnumber] != NULL)
			{
				++clientnumber;
			}
			printf("SA:\tGot a connection, clientnumber: %i\n", clientnumber);

			/* Tell them the max players so they can allocate their array! */
			sBuf[0] = 'M';
			sBuf[1] = 'P';
			sBuf[2] = clients->max;
			if(ssend(sock, sBuf, sizeof(avatar)) != sizeof(avatar))
			{
				printf("SA:\tERROR: Could not tell client the max number of players (%i)\n", clients->max);
				sclose(sock);
				continue;
			}

			/* Tell them their ID */
			sBuf[0] = 'I';
			sBuf[1] = 'D';
			sBuf[2] = clientnumber;
			if(ssend(sock, sBuf, sizeof(avatar)) != sizeof(avatar))
			{
				printf("SA:\tERROR: Could not tell client their ID (%i)\n", clientnumber);
				sclose(sock);
				continue;
			}

			/* Get the client and add them to the list */
			newClient = (client*)smalloc(sizeof(client));
			newClient->sock = sock;
			newClient->shapeChange = 1;
			newClient->leftDown = false;
			newClient->rightDown = false;
			clients->cs[clientnumber] = newClient;

			/* Same for their avatar */
			newAvatar = (avatar*)smalloc(sizeof(avatar));
			newAvatar->id = clientnumber;
			newAvatar->r = 0.0f;
			newAvatar->x = *(((int**)data)[4]);
			newAvatar->y = *(((int**)data)[5]);
			newAvatar->vx = 0.0f;
			newAvatar->vy = 0.0f;
			newAvatar->score = 0;
			newAvatar->shape = circle;
			newAvatar->popped = false;
			avatars->as[clientnumber] = newAvatar;

			/* Increment counters for clients and avatars */
			clients->cur++;
			avatars->cur++;

			printf("SA:\tSpawning a listener thread for the client\n");
			dCAC[1] = &clientnumber;
			clients->cs[clientnumber]->thread = SDL_CreateThread(handleClient, dCAC);
		}
	}

	printf("SA:\tServer's going down, accept thread gracefully dying\n");

	return SUCCESS;
}

int handleClient(void* data)
{
	const bool* restrict done = ((bool**)data)[0];
	const unsigned char clientnumber = *(((unsigned char**)data)[1]);
	clientList* clients = ((clientList**)data)[2];
	avatarList* restrict avatars = ((avatarList**)data)[3];

	/* Recv buffer */
	char buf[2];

	/* The current client we're dealing with */
	client* cc = clients->cs[clientnumber];

	assert(done != NULL);
	assert(clients != NULL);
	assert(clientnumber < clients->max);

	while(!*done)
	{
		if(srecv(cc->sock, buf, 2) != 2)
		{
			printf("HC(%i):\tCould not get data over socket\n", clientnumber);
			continue;
		}

		/* Change in the right arrow */
		if(*buf == 'R')
		{
			/* Right arrow is now pressed */
			if(buf[1] == 'D')
			{
				cc->rightDown = true;
				
			}

			/* Right arrow is now released */
			else
			{
				assert(buf[1] == 'U');

				cc->rightDown = false;
			}
		}

		/* Left arrow */
		else if(*buf == 'L')
		{
			if(buf[1] == 'D')
			{
				cc->leftDown = true;
			}
			else
			{
				assert(buf[1] == 'U');

				cc->leftDown = false;
			}
		}

		/* Trying to change shapes */
		else if(*buf == 'S')
		{
			assert(buf[1] >= '0');
			assert(buf[1] < '5');

			cc->shapeChange = atoi(buf+1);
		}

		/* Client is quitting */
		else if(*buf == 'E')
		{
			assert(buf[1] == 'O');

			printf("HC(%i):\tClient is quitting\n", clientnumber);
			free(cc);
			clients->cs[clientnumber] = NULL;
			clients->cur--;

			free(avatars->as[clientnumber]);
			avatars->as[clientnumber] = NULL;
			avatars->cur--;

			return SUCCESS;
		}

		/* This should never happen */
		else
		{
			assert(false);
		}
	}

	printf("HC(%i):\tServer's going down, client handler thread gracefully dying\n", clientnumber);

	return SUCCESS;
}
