#include "shapewars.h"

/* The screen we draw to */
SDL_Surface* restrict screen;

/* Stores the walk mesh information */
GLuint meshtiles;

/* Store the avatar textures. */
GLuint avatextures[4];

float angle = 1.0f;

static map* restrict m;
maptile tiles[4];

int screenPush(short int velocity)
{
	return MIN(10, velocity);
}

void setupTile(maptile* restrict mt, short int x, short int y)
{
	//TODO: Is this function actually working right?  I think not.
	
	if(mt->tex != 0)
		glDeleteTextures(1, &mt->tex);

	if(x < 0)
	{
		x += 2;
	}
	else if(x > m->maptx)
	{
		x -= 2;
	}

	if(y < 0)
	{
		y += 2;
	}
	else if(y > m->mapty)
	{
		y -= 2;
	}

	/* Copy over the name */
	strcpy(mt->filename, m->tiles[x][y]);

	/* Load the texture */
	makeTexture(m->tiles[x][y], &(mt->tex));
}
	
inline int makeTexture(const char* restrict filename, GLuint* storage)
{
	/* Temporary surface. */
	SDL_Surface* temp;

	/* Load our palette file. */
	temp = IMG_Load(filename);
	if(!temp)
	{
		fprintf(stderr, "makeTexture:\tCould not load image (%s).\n", filename);
		SDL_FreeSurface(temp);
		SDL_Quit();
		return IMGLOADFAIL;
	}

	/* Generate the texture and set up parameters and filtering modes. */
	glGenTextures(1, storage);
	glBindTexture(GL_TEXTURE_2D, *storage);
	glPixelStorei(GL_UNPACK_ROW_LENGTH,
		(temp->pitch / temp->format->BytesPerPixel));
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA8, temp->w, temp->h, 0,
		GL_RGBA, GL_UNSIGNED_BYTE, temp->pixels);
	SDL_FreeSurface(temp);

	return SUCCESS;
}

int initGL(int width, int height, int bpp, fileList* files, int flags)
{
	/* Maximum texture size supported. */
	GLint MAXTexSize;

	/* Set up OpenGL attributes. */
	if(bpp >= 24)
	{
		SDL_GL_SetAttribute(SDL_GL_RED_SIZE, 8);
		SDL_GL_SetAttribute(SDL_GL_GREEN_SIZE, 8);
		SDL_GL_SetAttribute(SDL_GL_BLUE_SIZE, 8);
	}
	else
	{
		SDL_GL_SetAttribute(SDL_GL_RED_SIZE, 5);
		SDL_GL_SetAttribute(SDL_GL_GREEN_SIZE, 5);
		SDL_GL_SetAttribute(SDL_GL_BLUE_SIZE, 5);
	}
	SDL_GL_SetAttribute(SDL_GL_DEPTH_SIZE, 16);
	SDL_GL_SetAttribute(SDL_GL_DOUBLEBUFFER, 1);
	SDL_GL_SetAttribute(SDL_GL_SWAP_CONTROL, 1);

	/* Init screen */
	if((screen = SDL_SetVideoMode(width, height, bpp, flags)) == NULL)
	{
		fprintf(stderr, "initGL:\tCould not initialize SDL/GL mode: %s\n", SDL_GetError());
		SDL_Quit();
		return SDL_SETVIDEOMODEFAIL;
	}

	/* Make sure we can support textures of the proper size. */
	glGetIntegerv(GL_MAX_TEXTURE_SIZE, &MAXTexSize);
	if(MAXTexSize < MAX(m->tilex, m->tiley))
	{
		fprintf(stderr, "initGL:\tNeed at least %ix%i textures.\n", m->tilex, m->tiley);
		SDL_Quit();
		return MAXTEXSIZEFAIL;
	}

	/* Load a texture for each avatar. */
	if(makeTexture(files->circleFile, &avatextures[0]))
	{
		fprintf(stderr, "initGL:\tFailed to load the circle texture (%s).\n", files->circleFile);
		SDL_Quit();
		return AVAFAIL;
	}

	if(makeTexture(files->triangleFile, &avatextures[1]))
	{
		fprintf(stderr, "initGL:\tFailed to load the triangle texture (%s).\n", files->triangleFile);
		SDL_Quit();
		return AVAFAIL;
	}

	if(makeTexture(files->pentagonFile, &avatextures[2]))
	{
		fprintf(stderr, "initGL:\tFailed to load the pentagon texture (%s).\n", files->pentagonFile);
		SDL_Quit();
		return AVAFAIL;
	}

	if(makeTexture(files->semicircleFile, &avatextures[3]))
	{
		fprintf(stderr, "initGL:\tFailed to load the semicircle texture (%s).\n", files->semicircleFile);
		SDL_Quit();
		return AVAFAIL;
	}

	/* Set background and drawing colors to black. */
	glClearColor(0.0f, 0.0f, 0.0f, 0.0f);
	glColor4f(0.0f, 0.0f, 0.0f, 0.0f);

	/* Set up OpenGL for 2D rendering. */
	glDisable(GL_DEPTH_TEST);
	glDisable(GL_CULL_FACE);
	glViewport(0, 0, screen->w, screen->h);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glOrtho(0, width/2, height/2, 0, -1.0, 1.0);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	glTranslatef(0.0f, 0.0f, 0.0f);

	return SUCCESS;
}

static void stateReset()
{
	glState.do_blend = false;
	glState.do_texture = false;
	glState.texture = 0;
	glState.sfactor = 0xffffffff;
	glState.dfactor = 0xffffffff;
}

static void doBlend(bool on)
{
	if(glState.do_blend == on)
	{
		return;
	}

	on ? glEnable(GL_BLEND) : glDisable(GL_BLEND);
	glState.do_blend = on;
}

static void doTexture(bool on)
{
	if(glState.do_texture == on)
	{
		return;
	}

	on ? glEnable(GL_TEXTURE_2D) : 	glDisable(GL_TEXTURE_2D);
	glState.do_texture = on;
}

static void setBlend(GLenum sfactor, GLenum dfactor)
{
	if(glState.sfactor == sfactor && glState.dfactor == dfactor)
	{
		return;
	}

	glBlendFunc(sfactor, dfactor);
	glState.sfactor = sfactor;
	glState.dfactor = dfactor;
}

static void setTexture(GLuint tex)
{
	if(glState.texture == tex)
	{
		return;
	}

	glBindTexture(GL_TEXTURE_2D, tex);
	glState.texture = tex;
}

void drawTile(const maptile* restrict t, short int top, short int bottom, short int right, short int left, const bool pu, const bool pr, const short int p)
{
	/* Where should we draw on the screen? */
	int stop, sbottom, sleft, sright;

	/* Where in the texture do we draw from? */
	float ttop, tbottom, tleft, tright;
	
	if((top == bottom) || (left == right))
		return;

	/* Determine where to draw on the screen. */
	switch(p)
	{
		case 0:
			stop = (pu ? (top - bottom) : screen->h) / 2;
			sbottom = (pu ? 0 : (screen->h - (top - bottom))) / 2;
			sleft = (pr ?  0 : (screen->w - (right - left))) / 2;
			sright = (pr ? (right - left) : screen->w) / 2;
			break;
		case 1:
			stop = (pu ? (top - bottom) : screen->h) / 2;
			sbottom = (pu ? 0 : (screen->h - (top - bottom))) / 2;
			sleft = (pr ? screen->w - (right - left) : 0) / 2;
			sright = (pr ? screen->w : (right - left)) / 2;
			break;
		case 2:
			stop = (pu ? screen->h : (top - bottom)) / 2;
			sbottom = (pu ? (screen->h - (top - bottom)): 0) / 2;
			sleft = (pr ? 0 : (screen->w - (right - left))) / 2;
			sright = (pr ? (right - left) : screen->w) / 2;
			break;
		case 3:
			stop = (pu ? screen->h : (top - bottom)) / 2;
			sbottom = (pu ? (screen->h - (top - bottom)): 0) / 2;
			sleft = (pr ? screen->w - (right - left) : 0) / 2;
			sright = (pr ? screen->w : (right - left)) / 2;
			break;
		default:
			stop = 0;
			sbottom = 0;
			sleft = 0;
			sright = 0;
			break;
	}

	if((sleft == sright) || (stop == sbottom))
		return;

	/* Convert numbers to the 0-1 range. */
	ttop = (float)top / m->tiley;
	tbottom = (float)bottom / m->tiley;
	tleft = (float)left / m->tilex;
	tright = (float)right / m->tilex;

	/* Set up texturing and make sure blending is off. */
	doTexture(true);
	doBlend(false);
	setTexture(t->tex);

	/* Draw the tile to the screen.
	 * 	Top-Left
	 * 	Top-Right
	 * 	Bottom-Right
	 * 	Bottom-Left
	 */
	glColor4ub(255, 255, 255, 255);
	glBegin(GL_QUADS);
		glTexCoord2f(tleft, ttop); glVertex2f(sleft, stop);
		glTexCoord2f(tright, ttop); glVertex2f(sright, stop);
		glTexCoord2f(tright, tbottom); glVertex2f(sright, sbottom);
		glTexCoord2f(tleft, tbottom); glVertex2f(sleft, sbottom);
	glEnd();
}

void fillRect(SDL_Rect* r, Uint32 color)
{
	int dx1, dy1;
	int dx2, dy2;

	doTexture(false);
	doBlend(false);

	if(!r)
	{
		r = &screen->clip_rect;
	}

	dx1 = r->x;
	dy1 = r->y;
	dx2 = dx1 + r->w;
	dy2 = dy1 + r->h;

	glBegin(GL_QUADS);
	glColor3ub((color>>16) & 255, (color>>8) & 255, color & 255);
	glVertex2i(dx1, dy1);
	glVertex2i(dx2, dy1);
	glVertex2i(dx2, dy2);
	glVertex2i(dx1, dy2);
	glEnd();
}

Uint32 mapRGB(Uint8 r, Uint8 g, Uint8 b)
{
	return r<<16 | g<<8 | b;
}

void drawAvatar(int xPos, int yPos, const int av)
{
	/* This gives us 64x64 avatars. */
	const int dist = 32;

	/* Where do we draw on the screen? */
	int sright, sleft, stop, sbottom;

	/* How much texture do we want to draw? */
	float tright, tleft, ttop, tbottom;

	/* We need to set up texturing and blending. */
	doTexture(true);
	doBlend(true);

	/* Set the texture to the avatar we're drawing.
	 * Set blending to make alpha-channel transparency work.
	 */
	setTexture(avatextures[av]);
	setBlend(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	
	/* Figure out where to draw the center of the avatar. */
	xPos += (screen->w)/2;
	yPos += (screen->h)/2;

	/* Determine where to draw. */
	sright = MIN(xPos + dist, screen->w) / 2;
	sleft = MAX(xPos - dist, 0) / 2;
	stop = MIN(yPos + dist, screen->h) / 2;
	sbottom = MAX(yPos - dist, 0) / 2;

	/* Determine what part of the texture we need. */
	tright = (xPos + dist <= screen->w) ? 1.0f : (128 - (((xPos + dist) - screen->w) * 2)) / 128;
	tleft  = (xPos - dist >= 0) ? 0.0f : (0 + ((0 - (xPos - dist)) * 2)) / 128;
	ttop = (yPos + dist <= screen->h) ? 1.0f : (128 - (((yPos + dist) - screen->h) * 2)) / 128;
	tbottom = (yPos - dist <= screen->h) ? 0.0f : (0 + ((0 - (yPos - dist)) * 2)) / 128;

	/* Draw the avatar.
	 * 	Top-Right
	 * 	Bottom-Right
	 * 	Bottom-Left
	 * 	Top-Left
	 */
	glBegin(GL_QUADS);
	glColor4ub(255, 255, 255, 255);
		glTexCoord2f(tright, ttop); glVertex2f(sright, stop);
		glTexCoord2f(tright, tbottom); glVertex2f(sright, sbottom);
		glTexCoord2f(tleft, tbottom); glVertex2f(sleft, sbottom);
		glTexCoord2f(tleft, ttop); glVertex2f(sleft, stop);
	glEnd();
}

void fixBounds(int* top, int* bottom, int* left, int* right)
{
	if(*left < 0) *left = 0;
	else if(*left > m->tilex) *left = m->tilex;

	if(*right < 0) *right = 0;
	else if(*right > m->tilex) *right = m->tilex;

	if(*top < 0) *top = 0;
	else if(*top > m->tiley) *top = m->tiley;

	if(*bottom < 0) *bottom = 0;
	else if(*bottom > m->tiley) *bottom = m->tiley;
}

void drawGLScene(const avatarList* restrict avatars)
{
	/* Get necessary idnformation. */
	const unsigned short int inx = (int)avatars->as[avatars->us]->x % m->tilex;
	const unsigned short int iny = (int)avatars->as[avatars->us]->y % m->tiley;
	const unsigned short int xPos = (int)avatars->as[avatars->us]->x / m->tilex;
	const unsigned short int yPos = (int)avatars->as[avatars->us]->y / m->tiley;
	const unsigned short int w2 = screen->w / 2;
	const unsigned short int h2 = screen->h / 2;
	const bool pr = (inx > (m->tilex / 2)) ? true : false;
	const bool pu = (iny > (m->tiley / 2)) ? true : false;
	const short int xp = screenPush(avatars->as[avatars->us]->vx);
	const short int yp = screenPush(avatars->as[avatars->us]->vy);

	/* Parameters for a texture piece. */
	int top, bottom, right, left;

	/* Store parameters of the first tile we drew. */
	int otop, obottom, oright, oleft;

	/* Loop variable. */
	int i;

	/* Clear the buffers. */
	glClear(GL_COLOR_BUFFER_BIT);

	/* Initialize top, bottom, left, right. */
	top = 0;
	bottom = 0;
	left = 0;
	right = 0;

	/* Draw the first tile. */
	oleft = inx - w2 + xp;
	oright = inx + w2 + xp;
	obottom = iny - h2 + yp;
	otop = iny + h2 + yp;
	fixBounds(&otop, &obottom, &oleft, &oright);
	drawTile(&tiles[0], otop, obottom, oright, oleft, pu, pr, 0);

	/* Draw the second tile. */
	if(((xPos % m->maptx != 0) && !pr)
		|| ((xPos % m->maptx != m->maptx-1) && pr))
	{
		right = (pr ? screen->w - (oright - oleft) : m->tilex);
		left = (pr ? 0 : m->tilex - screen->w + (oright - oleft));
		fixBounds(&otop, &obottom, &left, &right);
		drawTile(&tiles[1], otop, obottom, right, left, pu, pr, 1);
	}

	/* Draw the third tile. */
	if(((yPos % m->mapty != m->mapty-1) && pu)
		|| ((yPos % m->mapty != 0) && !pu))
	{
		top = (pu ? screen->h - (otop - obottom) : m->tiley);
		bottom = (pu ? 0 : m->tiley - screen->h + (otop - obottom));
		fixBounds(&top, &bottom, &oleft, &oright);
		drawTile(&tiles[2], top, bottom, oright, oleft, pu, pr, 2);
	}

	/* Draw the fourth tile. */
	fixBounds(&top, &bottom, &left, &right);
	drawTile(&tiles[3], top, bottom, right, left, pu, pr, 3);

	/* Draw the avatars that should be on our screen. */
	for(i = 0; i < avatars->cur; i++)
	{
		if(!avatars->as[i])
			continue;

		if(i != avatars->us)
		{
			if((abs(avatars->as[i]->x - avatars->as[avatars->us]->x) < w2)
				&& (abs(avatars->as[i]->y - avatars->as[avatars->us]->y < h2)))
			{
				drawAvatar((avatars->as[i]->x - avatars->as[avatars->us]->x),
					(avatars->as[i]->y - avatars->as[avatars->us]->y),
					avatars->as[i]->shape-1);
			}
		}
		else
		{
			drawAvatar(xp, yp, avatars->as[avatars->us]->shape-1);
		}
	}

	SDL_GL_SwapBuffers();
}

int main(int argc __attribute__((unused)), char** argv __attribute__((unused)))
{
	/* The array of all our avatar objects */
	avatarList avaList;

	/* Store the locations of our files. */
	fileList files;

	/* False until the user quits */
	bool done = false;

	/* Five-pointer array for passing done, avatars, sock and mutex stuff to thread funcs */
	void* doneAndAvatars[5];

	/* Two ticks trackers. */
	long tick1, tick2;

	/* Flags to pass to initGL() */
	int flags;

	/* Useless loop variables. */
	int i;

	/* Physics thread */
	SDL_Thread* restrict physicsThread = NULL;

	/* Input / Network-out thread */
	SDL_Thread* restrict inputThread = NULL;

	/* Network-in thread */
	SDL_mutex* avatarLock = NULL;
	SDL_mutex* goLock = NULL;
	SDL_Thread* restrict netInThread = NULL;

	/* Config file key-value pairs */
	kvlist* config;
	keyval* kv;

	/* Server and port */
	char* server;
	unsigned short int port;

	/* The socket, man */
	TCPsocket sock;

	/* Temp. variables for tile loading. */
	char* tempName;
	GLuint tempTex;

	/* Screen width, height, bpp -- set to defaults */
	unsigned short int width = 1024, height = 768, bpp = 24, fullScreen = 0;

	/* Buffer to when we get stuff from the server */
	char buf[sizeof(avatar)];

	/* TODO: comment*/
	unsigned short int tilex, tiley;
	bool left, down;

	/* Keeps track of delta tick. */
	float dt = 0.0f;

	/* Initialize map. */
	m = NULL;

	/* Allocate memory for files names. */
	files.meshFile = (char*)scalloc(32, sizeof(char));
	files.worldFile = (char*)scalloc(32, sizeof(char));
	files.circleFile = (char*)scalloc(32, sizeof(char));
	files.triangleFile = (char*)scalloc(32, sizeof(char));
	files.pentagonFile = (char*)scalloc(32, sizeof(char));
	files.semicircleFile = (char*)scalloc(32, sizeof(char));

	/* Set up default file names. */
	strcpy(files.meshFile, MESHFILE);
	strcpy(files.worldFile, WORLDFILE);
	strcpy(files.circleFile, CIRCLEFILE);
	strcpy(files.triangleFile, TRIANGLEFILE);
	strcpy(files.pentagonFile, PENTAGONFILE);
	strcpy(files.semicircleFile, SEMICIRCLEFILE);

	/* Allocate memory for server and set up defaults. */
	server = (char*)scalloc(128, sizeof(char));
	strcpy(server, "localhost");
	port = 55112;

	/* Get memory for our tempName and initialize tempTex. */
	tempName = (char*)calloc(128, sizeof(char));
	tempTex = 0;

	/* Init SDL */
	if(SDL_Init(SDL_INIT_TIMER | SDL_INIT_AUDIO | SDL_INIT_VIDEO) == -1)
	{
		fprintf(stderr, "Main:\tCould not SDL_Init()\n");
		return SDLINITFAIL;
	}

	/* Make sure we call SDL_Quit() upon exiting. */
	atexit(SDL_Quit);

	config = readConfigFile(CONFIGNAME);

	while(!kvlistIsEmpty(config))
	{
		kv = kvlistPop(config);
		if(!strcmp(kv->key, "Depth"))
		{
			bpp = atoi(kv->val);
		}
		else if(!strcmp(kv->key, "XRes"))
		{
			width = atoi(kv->val);
		}
		else if(!strcmp(kv->key, "YRes"))
		{
			height = atoi(kv->val);
		}
		else if(!strcmp(kv->key, "FullScreen"))
		{
			fullScreen = atoi(kv->val);
			if(fullScreen > 1)
			{
				fullScreen = 1;
			}
		}
		else if(!strcmp(kv->key, "Host"))
		{
			server = (char*)calloc(strlen(kv->val) + 1, sizeof(char));
			strcpy(server, kv->val);
		}
		else if(!strcmp(kv->key, "Port"))
		{
			port = atoi(kv->val);
		}
		else if(!strcmp(kv->key, "CircleFile"))
		{
			memset(files.circleFile, 0, strlen(files.circleFile));
			strcpy(files.circleFile, kv->val);
		}
		else if(!strcmp(kv->key, "TriangleFile"))
		{
			memset(files.triangleFile, 0, strlen(files.triangleFile));
			strcpy(files.triangleFile, kv->val);
		}
		else if(!strcmp(kv->key, "PentagonFile"))
		{
			memset(files.pentagonFile, 0, strlen(files.pentagonFile));
			strcpy(files.pentagonFile, kv->val);
		}
		else if(!strcmp(kv->key, "SemicircleFile"))
		{
			memset(files.semicircleFile, 0, strlen(files.semicircleFile));
			strcpy(files.semicircleFile, kv->val);
		}
		else if(!strcmp(kv->key, "MeshFile"))
		{
			memset(files.meshFile, 0, strlen(files.meshFile));
			strcpy(files.meshFile, kv->val);
		}
		else if(!strcmp(kv->key, "WorldFile"))
		{
			memset(files.worldFile, 0, strlen(files.worldFile));
			strcpy(files.worldFile, kv->val);
		}
		else if(!strcmp(kv->key, "Server"))
		{
			memset(server, 0, strlen(server));
			strcpy(server, kv->val);
		}
		else if(!strcmp(kv->key, "Port"))
		{
			port = atoi(kv->val);
		}

		sfree(kv->key);
		sfree(kv->val);
		kvalDestroy(kv);
	}

	kvlistDestroy(config);

	/* Set up flags. */
	flags = SDL_OPENGL;
	if(fullScreen)
		flags |= SDL_FULLSCREEN;

	/* Ensure that the desired video mode is okay before proceeding. */
	fprintf(stdout, "width %i, height %i, bpp %i\n", width, height, bpp);
	if(SDL_VideoModeOK(width, height, bpp, flags))
	{
		fprintf(stdout, "Video mode OK.\n");
	}
	else
	{
		fprintf(stderr, "Video mode not OK.\n");
		SDL_Quit();
		return SDL_VIDEOMODEOKFAIL;
	}

	/* Load the first map. */ // TODO: We gots to get this from the server, yo!
	assert(files.worldFile != NULL);
	m = loadMap(files.worldFile);
	if(!m)
	{
		fprintf(stderr, "Main:\tFailed to load the first level. (%s)\n", files.worldFile);
		return MAPFAIL;
	}

	/* Start with zero players */
	avaList.max = 0;
	avaList.cur = 0;

	/* Let's do the stuff here, yeah */
	/*
		Here's what we got to do:
		1) Connect to the server
		2) Get our MaxPlayers then ID
		3) Get the avatars and shit (let's do this in the netin thread, yeah)
		4) Aww, yeah.
	*/

	/* Get our connection */
	if(!(sock = socketConnect(server, port)))
	{
		fprintf(stderr, "Main:\tFailed to connect to host %s for port %u\n", server, port);
		return 0xBAD;
	}

	/* All message are sizeof(avatar) long. */
	srecv(sock, buf, sizeof(avatar));

	/* So what's the first message that we get?  Let's check protocol.spec. */
	/* It's our MaxPlayers, alright */

	/* But does the server play along? */
	if(*buf != 'M' || *(buf+1) != 'P')
	{
		fprintf(stderr, "Main:\tWhoa, man.  The server is playin' us for chumps on the MP front!!\n");
		return MPFAIL;
	}

	/* We're good */
	else
	{
		avaList.as = (avatar**)calloc(buf[2], sizeof(avatar*));
		avaList.max = buf[2];
	}

	/* Let's get our ID now */
	srecv(sock, buf, sizeof(avatar));

	/* Make sure the server's down too */
	if(*buf != 'I' || *(buf + 1) != 'D')
	{
		fprintf(stderr, "Main:\tWhoa, man.  The server is playin' us for chumps on the ID front!\n");
		return IDFAIL;
	}

	/* We're cool */
	else
	{
		/* Got our ID!  Booyah. */
		avaList.us = *(buf+2);
	}
	/* Alright, we're good to let the netin thread take over now */

	/* Initialize OpenGL */
	if(initGL(width, height, bpp, &files, flags))
	{
		fprintf(stderr, "Main:\tFailed to initialize OpenGL.\n");
		return INITGL_FAIL;
	}

	/* Set window caption and turn off mouse cursor display. */
	SDL_WM_SetCaption(GAMENAME, NULL);
	SDL_ShowCursor(0);

	/* Initialize our glState. */
	stateReset();

	/* Initialize mutex stuff */
	goLock = SDL_CreateMutex();
	avatarLock = SDL_CreateMutex();

	/* Initialize doneAndAvatars. */
	doneAndAvatars[0] = &done;
	doneAndAvatars[1] = &avaList;
	doneAndAvatars[2] = &sock;
	doneAndAvatars[3] = &goLock;
	doneAndAvatars[4] = &avatarLock;

	/* Set up threads. */
	physicsThread = SDL_CreateThread(physics, doneAndAvatars);
	inputThread = SDL_CreateThread(input, doneAndAvatars);
	netInThread = SDL_CreateThread(netIn, doneAndAvatars);

	/* Initialize the tick tracker. */
	tick1 = SDL_GetTicks();

	/* Wait here until avatars has been set up from the server */
	sleep(1);
	SDL_LockMutex(goLock);

	/* Set up our initial tiles.  The tile array is:
	 *  0: Current Tile
	 *  1: Left/Right Tile
	 *  2: Up/Down Tile
	 *  3: Corner Tile
	 */
	SDL_LockMutex(avatarLock);
	tilex = avaList.as[avaList.us]->x / m->tilex;
	tiley = avaList.as[avaList.us]->y / m->tiley;
	left = (((int)avaList.as[avaList.us]->x % m->tilex) < (m->tilex / 2));
	down = (((int)avaList.as[avaList.us]->y % m->tiley) < (m->tiley / 2));
	SDL_UnlockMutex(avatarLock);
	setupTile(&(tiles[0]),
		tilex,
		tiley);
	setupTile(&(tiles[1]),
		tilex + (((tilex == 0 && left) || (tilex+1 == m->maptx && !left)) ? -1 : 1) * ((left) ? -1 : 1),
		tiley);
	setupTile(&(tiles[2]),
		tilex,
		(((tiley == 0 && down) || (tiley+1 == m->mapty && !down)) ? -1 : 1) * ((down) ? -1 : 1));
	setupTile(&(tiles[3]),
		tilex + (((tilex == 0 && left) || (tilex+1 == m->maptx && !left)) ? -1 : 1) * ((left) ? -1 : 1),
		(((tiley == 0 && down) || (tiley+1 == m->mapty && !down)) ? -1 : 1) * ((down) ? -1 : 1));

	/* Graphics / Sound */
	while(!done)
	{
		/* Figure out which tile we're in and whether we're pushing up or down and left or right. */
		SDL_LockMutex(avatarLock);
		tilex = avaList.as[avaList.us]->x / m->tilex;
		tiley = avaList.as[avaList.us]->y / m->tiley;
		left = (((int)avaList.as[avaList.us]->x % m->tilex) < (m->tilex / 2));
		down = (((int)avaList.as[avaList.us]->y % m->tiley) < (m->tiley / 2));

		/* If we're not in the same tile that we were last time, we need to load new tiles. */
		if(strcmp(tiles[0].filename, m->tiles[tilex][tiley]))
		{
			/* We're now in the left/right tile. */
			if(!strcmp(tiles[1].filename, m->tiles[tilex][tiley]))
			{
				/* Switch our current tile with the left/right tile. */
				strcpy(tempName, tiles[0].filename);
				tempTex = tiles[0].tex;

				strcpy(tiles[0].filename, tiles[1].filename);
				tiles[0].tex = tiles[1].tex;
				
				strcpy(tiles[1].filename, tempName);
				tiles[1].tex = tempTex;

				/* Switch the up/down tile with the corner tile. */
				strcpy(tempName, tiles[2].filename);
				tempTex = tiles[2].tex;

				strcpy(tiles[2].filename, tiles[3].filename);
				tiles[2].tex = tiles[3].tex;

				strcpy(tiles[3].filename, tempName);
				tiles[3].tex = tempTex;
			}
			/* We're now in the up/down tile. */
			else if(!strcmp(tiles[2].filename, m->tiles[tilex][tiley]))
			{
				/* Switch our current tile and the up/down tile. */
				strcpy(tempName, tiles[0].filename);
				tempTex = tiles[0].tex;

				strcpy(tiles[0].filename, tiles[2].filename);
				tiles[0].tex = tiles[2].tex;

				strcpy(tiles[2].filename, tempName);
				tiles[2].tex = tempTex;

				/* Switch our left/right tile and the corner tile. */
				strcpy(tempName, tiles[1].filename);
				tempTex = tiles[1].tex;

				strcpy(tiles[1].filename, tiles[3].filename);
				tiles[1].tex = tiles[3].tex;

				strcpy(tiles[3].filename, tempName);
				tiles[3].tex = tempTex;
			}
			/* We're now in the corner tile. */
			else if(!strcmp(tiles[3].filename, m->tiles[tilex][tiley]))
			{
				/* Switch our current tile with the corner tile. */
				strcpy(tempName, tiles[0].filename);
				tempTex = tiles[0].tex;
		
				strcpy(tiles[0].filename, tiles[3].filename);
				tiles[0].tex = tiles[3].tex;
		
				strcpy(tiles[3].filename, tempName);
				tiles[3].tex = tempTex;
		
				/* Switch the up/down tile with the left/right tile. */
				strcpy(tempName, tiles[1].filename);
				tempTex = tiles[1].tex;
		
				strcpy(tiles[1].filename, tiles[2].filename);
				tiles[1].tex = tiles[2].tex;
		
				strcpy(tiles[2].filename, tempName);
				tiles[2].tex = tempTex;
			}
			/* We're somehow skipped all of the surrounding tiles, so we need to
			 * figure everything out again.
			 */
			else
			{
				fprintf(stderr, "crazy teleport code!\n");
				setupTile(&(tiles[0]),
					 tilex,
					 tiley);
				setupTile(&(tiles[1]),
					tilex + (((tilex == 0 && left) || (tilex+1 == m->maptx && !left)) ? -1 : 1) * ((left) ? -1 : 1),
					tiley);
				setupTile(&(tiles[2]),
					tilex,
					(((tiley == 0 && down) || (tiley+1 == m->mapty && !down)) ? -1 : 1) * ((down) ? -1 : 1));
				setupTile(&(tiles[3]),
					tilex + (((tilex == 0 && left) || (tilex+1 == m->maptx && !left)) ? -1 : 1) * ((left) ? -1 : 1),
					(((tiley == 0 && down) || (tiley+1 == m->mapty && !down)) ? -1 : 1) * ((down) ? -1 : 1));
			}
		}
		/* We were pushing left/right but now we're pushing the opposite way. */
		if(strcmp(tiles[1].filename,
			m->tiles[tilex + (((tilex == 0 && left) || (tilex+1 == m->maptx && !left)) ? -1 : 1) * ((left) ? -1 : 1)][tiley]))
		{
			/* We need a new 1 and a new 3. */
			setupTile(&(tiles[1]),
				tilex + (((tilex == 0 && left) || (tilex+1 == m->maptx && !left)) ? -1 : 1) * ((left) ? -1 : 1),
				tiley);
			setupTile(&(tiles[3]),
				tilex + (((tilex == 0 && left) || (tilex+1 == m->maptx && !left)) ? -1 : 1) * ((left) ? -1 : 1),
				tiley + (((tiley == 0 && down) || (tiley+1 == m->mapty && !down)) ? -1 : 1) * ((down) ? -1 : 1));
		}
		/* We were pushing up/down, but now we're pushing the opposite way. */
		else if(strcmp(tiles[2].filename,
			m->tiles[tilex][tiley + (((tiley == 0 && down) || (tiley+1 == m->mapty && !down)) ? -1 : 1) * ((down) ? -1 : 1)]))
		{
			/* We need a new 2 and a new 3. */
			setupTile(&(tiles[2]),
				tilex,
				tiley + (((tiley == 0 && down) || (tiley+1 == m->mapty && !down)) ? -1 : 1) * ((down) ? -1 : 1));
			setupTile(&(tiles[3]),
				tilex + (((tilex == 0 && left) || (tilex+1 == m->maptx && !left)) ? -1 : 1) * ((left) ? -1 : 1),
				tiley + (((tiley == 0 && down) || (tiley+1 == m->mapty && !down)) ? -1 : 1) * ((down) ? -1 : 1));
		}
		
		//fprintf(stderr, "0: %s, 1: %s, 2: %s, 3: %s\n", tiles[0].filename, tiles[1].filename, tiles[2].filename, tiles[3].filename);
		/* Draw the scene */
		drawGLScene(&avaList);

		SDL_UnlockMutex(avatarLock);
	}

	/* Wait for threads to finish. */
	fprintf(stderr, "Waiting for the physics thread...\n");
	SDL_WaitThread(physicsThread, NULL);
	fprintf(stderr, "Waiting for the input thread...\n");
	SDL_WaitThread(inputThread, NULL);
	fprintf(stderr, "Waiting for the net thread...\n");
	SDL_WaitThread(netInThread, NULL);
	fprintf(stderr, "Exiting %s...\n", GAMENAME);

	/* Free mutex stuff */
	SDL_DestroyMutex(avatarLock);
	SDL_DestroyMutex(goLock);

	/* Free textures and clean up SDL. */
	glDeleteTextures(1, &meshtiles);
	glDeleteTextures(4, avatextures);
	for(i=0; i<4; i++)
	{
		glDeleteTextures(1, &(tiles[i].tex));
	}
	SDL_FreeSurface(screen);
	freeMap(m);
	sfree(avaList.as);
	sfree(server);
	sfree(files.meshFile);
	sfree(files.worldFile);
	sfree(files.circleFile);
	sfree(files.triangleFile);
	sfree(files.pentagonFile);
	sfree(files.semicircleFile);
	sfree(tempName);
	SDL_Quit();

	/* Exit. */
	return SUCCESS;
}
