/* SDL stuff */
#include <SDL/SDL.h>
#include <SDL/SDL_thread.h>
#include "SDL/SDL_image.h"

/* OpenGL stuff */
#include <GL/gl.h>
#include <GL/glu.h>

/* Console I/O */
#include <stdio.h>

/* Asserts are good */
#include <assert.h>

/* Key-value stuff */
#include "keyval.h"
#include "kvlist.h"

/* Strcmp() et cetera */
#include <string.h>

/* Avatars */
#include "avatar.h"

/* Threads */
#include "physics.h"
#include "input.h"
#include "netin.h"

/* Sleep */
#include <unistd.h>

/* Configuration options */
#include "config.h"

/* For floor() */
#include <math.h>

/* Networkin' shit */
#include "networking.h"

/* Map stuff */
#include "map.h"

#ifndef SHAPEWARS_H
#define SHAPEWARS_H

/* Filename for the config file. */
#define CONFIGNAME "shapewars.conf"

/* This is used as the window caption */
#define GAMENAME "Shape Wars"

/* Default filenames */
#define PALETTEFILE "maps/palette.png\0"
#define MESHFILE "maps/mesh.png\0"
#define WORLDFILE "maps/level1.map\0"
#define CIRCLEFILE "avatars/circle.png\0"
#define TRIANGLEFILE "avatars/triangle.png\0"
#define PENTAGONFILE "avatars/pentagon.png\0"
#define SEMICIRCLEFILE "avatars/semicircle.png\0"

/* Main() error codes */
#define SDL_VIDEOMODEOKFAIL 1
#define SDLINITFAIL 2
#define MAPFAIL 3
#define INITGL_FAIL 4
#define MPFAIL 5
#define IDFAIL 6

/* makeTexture() error codes */
#define IMGLOADFAIL 7

/* initGL() error codes */
#define SDL_SETVIDEOMODEFAIL 8
#define MAXTEXSIZEFAIL 9
#define PALETTEFAIL 10
#define MESHFAIL 11
#define AVAFAIL 12

/* Map function error codes. */
#define OPENFAIL 13
#define WRITEFAIL 14

/* For optimizing OpenGL states. */
static struct
{
	bool	do_blend;
	bool	do_texture;
	GLuint	texture;
	GLenum	sfactor, dfactor;
} glState;

/* For storing file names. */
typedef struct FileList
{
	char*	meshFile;
	char*	worldFile;
	char*	circleFile;
	char*	triangleFile;
	char*	pentagonFile;
	char*	semicircleFile;
} fileList;

SDL_mutex* goLock;

/* Makes a texture from an image. */
inline int makeTexture(const char* restrict filename, GLuint* storage);

/* Initalizes OpenGL */
int initGL(int width, int height, int bpp, fileList* files, int flags);

/* For changing the glState struct. */
static void stateReset(void);
static void doBlend(bool on);
static void doTexture(bool on);
static void setBlend(GLenum sfactor, GLenum dfactor);
static void setTexture(GLuint tx);

/* Maps three 8-bit values into one 32-bit value */
Uint32 mapRGB(Uint8 r, Uint8 g, Uint8 b);

/* Draws a tile to the screen. */
void drawTile(const maptile* restrict t, short int top, short int bottom, short int right, short int left, const bool pu, const bool pr, const short int p) __attribute__((nonnull));

/* Fills a rectangle on the screen. */
void fillRect(SDL_Rect* r, Uint32 color);

void drawAvatar(const int xPos, const int yPos, const int av);

/* Draw a scene. */
void drawGLScene(const avatarList* restrict avatars);

/* Makes sure an avatar stays within the world limits. */
void videoLimit(int width, int height);

#endif /* SHAPEWARS_H */
