/* Assertions are good */
#include <assert.h>

/* Malloc and such */
#include <stdlib.h>

#ifndef SHARED_H
#define SHARED_H

/* Global return codes */
#define SUCCESS 0

/* Bool defines */
typedef unsigned char bool;
#define true 1
#define false 0

/* Min/max macros, because they're useful. */
#define MAX(a,b) ((a > b) ? a : b)
#define MIN(a,b) ((a < b) ? a : b)

/**
 * __builtin_expect macros
 */
#define likely(x)       __builtin_expect((x), 1)
#define unlikely(x)     __builtin_expect((x), 0)

/**
 * Allocates some memory, like malloc().  Also asserts that it does not return
 * a null pointer.
 *
 * @param size The number of bytes to allocate
 * @return A pointer to the allocated memory
 */
inline void* smalloc(size_t size) __attribute__((malloc));

/**
 * Allocates some memory, like calloc().  Also asserts that it does not return
 * a null pointer.
 *
 * @param nelem The number of elements to allocate
 * @param size The number of bytes to allocate
 * @return A pointer to the allocated memory
 */
inline void* scalloc(size_t nelem, size_t size) __attribute__((malloc));

/**
 * Frees some memory, like free().  Also asserts that it was not input a null
 * pointer.
 *
 * @param ptr A pointer to the memory to be freed
 */
inline void sfree(void* restrict ptr);

#endif /*SHARED_H */
